unit Graph2DFunctions;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graph2D, multidim, mathConfig, math;

type
  {
   TGraphFunction
   draws one general function, given by fkt. Tries to buffer the function curve
   to avoid unnecessary re-computing of the function curve.
  }
  TGraphFunction = class(TGraph2D.TRenderObj)
   type
    SMD = specialize Multidimensional<MD1d, MD1d>;
   private
    buff : TGraph2D.TFunctionBuff;
    _fkt : SMD.IFunction;
    procedure writeFkt(f : SMD.IFunction); inline;
   protected
    procedure render; override;
   public
    name : String;
    style : TGraph2D.TLineStyle;
    constructor create (_Graph : TGraph2D);
    property fkt : SMD.IFunction read _fkt write writeFkt;
    procedure clearBuff;
    destructor Destroy; override;
  end;

implementation

constructor TGraphFunction.Create (_Graph : TGraph2D);
begin
 inherited Create(_Graph);
 _fkt := nil;
 Style := SOLID;
 Name := '';
 Buff := nil;
end;

procedure TGraphFunction.clearBuff;
begin
 SetLength(Buff, 0);
end;

procedure TGraphFunction.Render;
var
  b : Boolean;
  x, i : integer;
  f : Precision;
  accuracyDelta : Precision;
begin
  if (fkt = nil) then exit;
  if (graph.width = 0) then exit;
  accuracyDelta := (xAxis.screenToPos(1)-xAxis.screenToPos(0)) / 8;
  Graph.Canvas.Pen.Color := YAxis.Color;
  b := true;
  i := length(buff);
  if (i <> graph.width) then begin // resize buffer
    SetLength(buff, graph.width);
    for i := i to graph.width-1 do buff[i].x := 0/0;
  end else if (not isNaN(buff[0].x)) and (not graph.completeRepaint) then begin // move buffer to avoid re-calculation
    f := xAxis.ScreenToPos(0) - buff[0].x;
    if (abs((xAxis.ScreenToPos(graph.width-1) - buff[graph.width-1].x) - f) < abs(xAxis.ScreenToPos(1) - xAxis.ScreenToPos(0))) then begin
      i := round(f / (xAxis.ScreenToPos(1) - xAxis.ScreenToPos(0)));
      if (i > 0) then
        for x := 0 to graph.width-1-i do buff[x] := buff[x+i]
      else if (i < 0) then
        for x := graph.width-1 downto -i do buff[x] := buff[x+i];
    end;
  end;
  for x := 0 to Graph.Width - 1 do begin
    f := xAxis.ScreenToPos(x);
    if (isNaN(buff[x].x)) or (abs(buff[x].x - f) > accuracyDelta) or (graph.completeRepaint) then begin
      Buff[x].x := f;
      Buff[x].y := Fkt.Calc(f);
    end;
    if (isNaN(Buff[x].y)) then b := true
    else begin
      if (Buff[x].y < YAxis.FktMin) then YAxis.FktMin := Buff[x].y;
      if (Buff[x].y > YAxis.FktMax) then YAxis.FktMax := Buff[x].y;
      if (b) then begin Graph.MoveTo(x, YAxis.PosToScreen(Buff[x].y)); b := false; end
      else Graph.StyledLineTo(x, YAxis.PosToScreen(Buff[x].y), Style);
    end;
  end;
end;

procedure TGraphFunction.writeFkt(f: SMD.IFunction);
begin
 SetLength(Buff, 0);
 _Fkt := f;
end;

destructor TGraphFunction.Destroy;
begin
 inherited;
end;


end.

