unit interpolation2d;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, PairSplitter, ExtCtrls, Grids, StdCtrls,
  guitabbase, xmlUtil, Graph2D, heightmap, multidim, dialogs, ComCtrls,
  mathconfig, splines;

type

  { TInterpolation2Dim }

  TInterpolation2Dim = class(TGuiTab)
     Button1: TButton;
     Button2: TButton;
     Button3: TButton;
     Button4: TButton;
     Button5: TButton;
     CheckBox1: TCheckBox;
     CheckBox2: TCheckBox;
     ComboBox1: TComboBox;
     Label1: TLabel;
     Label2: TLabel;
     Label3: TLabel;
     Label4: TLabel;
     Label5: TLabel;
     PairSplitter1: TPairSplitter;
     PairSplitterSide1: TPairSplitterSide;
     PairSplitterSide2: TPairSplitterSide;
     ScrollBox1: TScrollBox;
     StringGrid1: TStringGrid;
     TrackBar1: TTrackBar;
     TrackBar2: TTrackBar;
     procedure addX(Sender: TObject);
     procedure addY(Sender: TObject);
     procedure bordaryChange(Sender: TObject);
     procedure ChangeDerivation(Sender: TObject);
     procedure ChangeGridVisibility(Sender: TObject);
     procedure ChangeSync(Sender: TObject);
     procedure delX(Sender: TObject);
     procedure delY(Sender: TObject);
     procedure genRandom(Sender: TObject);
     procedure graphsResize(Sender: TObject);
     procedure tableChange(Sender: TObject);
     procedure updateTable(Sender: TObject);
     procedure updateSpline(Sender: TObject);
   type
    SMD = specialize Multidimensional<MD2d, MD1d>;

    { TDerivationPaint }

    TDerivationPaint = class(SMD.IFunction)
     private
      fkt : Spline.TBicubicC2;
      dx, dy : Byte;
      function asObject : TObject;
     public
      constructor create;
      function calc(const x:MD2d) : MD1d;
      destructor destroy; override;
    end;
   private
    graphDat, graphSpl : TGraph2D;
    buf : SMD.TBuffer;
    bufMap : THeightMap;
    splMap : THeightMap;
    bufGrid : TGridPaint;
    splGrid : TGridPaint;
    class var instanceID : Cardinal;
   public
    constructor create(theOwner:TControl; __xmlEl:XML.TElement); override;
    destructor destroy; override;
  end;

implementation

{$R *.lfm}

{ TInterpolation2Dim.TDerivationPaint }

constructor TInterpolation2Dim.TDerivationPaint.create;
begin
  inherited;
  fkt := nil;
  dx := 0;
  dy := 0;
end;

function TInterpolation2Dim.TDerivationPaint.asObject: TObject;
begin
  result := self;
end;

function TInterpolation2Dim.TDerivationPaint.calc(const x: MD2d): MD1d;
begin
  result := fkt.derivative(dx, dy, x[0], x[1]);
end;

destructor TInterpolation2Dim.TDerivationPaint.destroy;
begin
  fkt.free;
  inherited;
end;

{ TInterpolation2Dim }

constructor TInterpolation2Dim.create(theOwner: TControl; __xmlEl: XML.TElement);
procedure loadData;
var
  str : TMemoryStream;
  s : String;
begin
  str := TMemoryStream.create;
  s := xmlEl.getTextContent;
  if (length(s) > 1) then str.write(s[2], length(s)); // also cutoff initial #
  str.position := 0;
  try if (s <> '') then StringGrid1.LoadFromCSVStream(str, #9, true, 0, false); except end;
  str.destroy;
  if (StringGrid1.rowCount > 1) and (StringGrid1.cells[0, StringGrid1.rowCount-1] = '') then StringGrid1.rowCount := StringGrid1.rowCount-1;
  if (StringGrid1.rowCount > 1) and (StringGrid1.colCount > 1) then begin
    try
      graphDat.xAxis[0].min := strToFloat(StringGrid1.cells[1, 0], mathFloatFormat);
      graphDat.xAxis[0].max := strToFloat(StringGrid1.cells[StringGrid1.ColCount-1, 0], mathFloatFormat);
      graphDat.yAxis[0].max := strToFloat(StringGrid1.cells[0, 1], mathFloatFormat);
      graphDat.yAxis[0].min := strToFloat(StringGrid1.cells[0, StringGrid1.RowCount-1], mathFloatFormat);
    except
    end;
    tableChange(self);
  end;
end;
begin
  inherited create(theOwner, __xmlEl);
  THeightMap.multithreading := true;
  name := 'Interpolation2Dim' + inttostr(instanceID);
  instanceID += 1;
  graphDat := TGraph2D.create(self);
  graphDat.parent := PairSplitterSide1;
  graphDat.anchorParallel(akTop, 0, PairSplitterSide1);
  graphDat.anchorParallel(akLeft, 0, PairSplitterSide1);
  graphDat.anchorParallel(akBottom, 0, PairSplitterSide1);
  graphSpl := TGraph2D.create(self);
  graphSpl.parent := PairSplitterSide1;
  graphSpl.anchorParallel(akTop, 0, PairSplitterSide1);
  graphSpl.anchorParallel(akRight, 0, PairSplitterSide1);
  graphSpl.anchorParallel(akBottom, 0, PairSplitterSide1);
  graphSpl.anchorToNeighbour(akLeft, 0, graphDat);
  graphSpl.synchronize(graphDat);
  buf := SMD.TBuffer.create;
  bufMap := THeightMap.create(graphDat);
  bufMap.fkt := buf;
  bufGrid := TGridPaint.create(graphDat);
  bufGrid.editable := true;
  bufGrid.grid := buf.grid;
  bufGrid.onChange := @updateTable;
  splMap := THeightMap.create(graphSpl);
  splMap.synchronize(bufMap);
  splGrid := TGridPaint.create(graphSpl);
  splGrid.grid := buf.grid;
  loadData;
end;

procedure TInterpolation2Dim.graphsResize(Sender: TObject);
begin
  graphDat.width := PairSplitterSide1.Width div 2;
end;

procedure TInterpolation2Dim.tableChange(Sender: TObject);
var
  grid : SMD.TGrid;
  ix, iy : integer;
  p : SMD.TGridPoint;
begin
  grid.init(2);
  for ix := 1 to StringGrid1.ColCount-1 do grid.addGridLane(0, StrToFloat(StringGrid1.Cells[ix, 0], mathFloatFormat));
  for iy := 1 to StringGrid1.RowCount-1 do grid.addGridLane(1, StrToFloat(StringGrid1.Cells[0, iy], mathFloatFormat));
  buf.setGrid(grid);
  for iy := 0 to buf.grid^.laneCount[1]-1 do for ix := 0 to buf.grid^.laneCount[0]-1 do begin
    p[0] := ix;
    p[1] := iy;
    try
      buf[p] := StrToFloat(StringGrid1.Cells[ix+1, buf.grid^.laneCount[1] - iy], mathFloatFormat);
    except
      buf[p] := 0;
      StringGrid1.Cells[ix+1, buf.grid^.laneCount[1] - iy] := '';
    end;
  end;
  grid.clear;
  if (sender = self) then graphDat.invalidate
  else graphDat.doPaint(sender);
  updateSpline(Sender);
end;

procedure TInterpolation2Dim.updateTable(Sender: TObject);
var
  ix, iy : integer;
  p : SMD.TGridPoint;
begin
  StringGrid1.ColCount := buf.grid^.laneCount[0]+1;
  StringGrid1.RowCount := buf.grid^.laneCount[1]+1;
  for ix := 0 to buf.grid^.laneCount[0]-1 do StringGrid1.Cells[ix+1, 0] := floattostr(buf.grid^[0, ix], mathFloatFormat);
  for iy := 0 to buf.grid^.laneCount[1]-1 do StringGrid1.Cells[0, iy+1] := floattostr(buf.grid^[1, buf.grid^.laneCount[1]-iy-1], mathFloatFormat);
  for iy := 0 to buf.grid^.laneCount[1]-1 do for ix := 0 to buf.grid^.laneCount[0]-1 do begin
    p[0] := ix;
    p[1] := iy;
    StringGrid1.Cells[ix+1, buf.grid^.laneCount[1] - iy] := floattostr(buf[p], mathFloatFormat);
  end;
  if (Sender = bufGrid) then updateSpline(Sender);
end;

procedure TInterpolation2Dim.updateSpline(Sender: TObject);
procedure saveData;
var
  str : TMemoryStream;
  s : String;
  i : integer;
begin
  str := TMemoryStream.create;
  StringGrid1.SaveToCSVStream(str, #9, true, false);
  setLength(s, str.Size);
  str.Position := 0;
  if (length(s) > 0) then str.Read(s[1], str.Size);
  str.destroy;
  for i := 1 to length(s) do if (s[i] = #10) and ((i = 1) or (s[i-1] <> #13)) then s[i] := #13;
  xmlEl.setTextContent('#'+s);
end;

begin
  if {(Sender = ComboBox) or } (splMap.fkt = nil) then begin
    if (splMap.fkt <> nil) then splMap.fkt.asObject.destroy;
    splMap.fkt := TDerivationPaint.create;
    TDerivationPaint(splMap.fkt.asObject).dx := TrackBar1.position;
    TDerivationPaint(splMap.fkt.asObject).dy := TrackBar2.position;
    TDerivationPaint(splMap.fkt.asObject).fkt := Spline.TBicubicC2.create;
  end;
  with (TDerivationPaint(splMap.fkt.asObject).fkt) do begin
    case (ComboBox1.ItemIndex) of
     0:bordaryCondition := NATURAL;
     1:bordaryCondition := HERMITE;
     2:bordaryCondition := NOT_A_KNOT;
     else raise Exception.create('Unknown bordary!');
    end;
    try computeFor(buf, buf.grid^); except end;
  end;
  if (sender = self) then graphSpl.invalidate
  else begin
    graphSpl.doPaint(Sender);
    saveData;
  end;
end;

procedure TInterpolation2Dim.genRandom(Sender: TObject);
var
  i : integer;
  str1, str2, s : string;
  x, y : integer;
  grid : SMD.TGrid;
  p : SMD.TGridPoint;
begin
  s := InputBox('Dimensions', 'enter the size of the random buffer', '10x10');
  str1 := '';
  str2 := 'x';
  for i := 1 to length(s) do
    if (s[i] = ' ') then continue
    else if (s[i] = 'x') then str2 := ''
    else if (str2 = 'x') then str1 += s[i]
    else str2 += s[i];
  try
    x := StrToInt(str1);
    y := StrToInt(str2);
  except
    ShowMessage('Invalid integer!');
  end;
  grid.init(2);
  for i := 0 to x-1 do grid.addGridLane(0, i);
  for i := 0 to y-1 do grid.addGridLane(1, -i);
  buf.setGrid(grid, false);
  randomize;
  p[1] := 0;
  while (p[1] < y) do begin
    p[0] := 0;
    while (p[0] < x) do begin
      buf[p] := trunc(random*1000) / 1000;
      p[0] := p[0] + 1;
    end;
    p[1] := p[1] + 1;
  end;
  graphDat.xAxis[0].min := -0.5;
  graphDat.xAxis[0].max := x-0.5;
  graphDat.yAxis[0].max := 0.5;
  graphDat.yAxis[0].min := -y+0.5;
  grid.clear;
  updateTable(Sender);
  graphDat.doPaint(self);
  updateSpline(Sender);
end;

procedure TInterpolation2Dim.addX(Sender: TObject);
var i : integer;
begin
  i := StringGrid1.ColCount;
  StringGrid1.ColCount := i+1;;
  if (i < 2) then StringGrid1.Cells[i, 0] := '0'
  else StringGrid1.Cells[i, 0] := FloatToStr(StrToFloat(StringGrid1.Cells[i-1, 0], mathFloatFormat)+1, mathFloatFormat);
  StringGrid1.FixedCols := 1;
  tableChange(Sender);
end;

procedure TInterpolation2Dim.addY(Sender: TObject);
var i : integer;
begin
  i := StringGrid1.RowCount;
  StringGrid1.InsertColRow(false, i);
  if (i < 2) then StringGrid1.Cells[0, i] := '0'
  else StringGrid1.Cells[0, i] := FloatToStr(StrToFloat(StringGrid1.Cells[0, i-1], mathFloatFormat)-1, mathFloatFormat);
  StringGrid1.FixedRows := 1;
  tableChange(Sender);
end;

procedure TInterpolation2Dim.bordaryChange(Sender: TObject);
begin

end;

procedure TInterpolation2Dim.ChangeDerivation(Sender: TObject);
begin
  if (splMap.fkt = nil) then exit;
  TDerivationPaint(splMap.fkt.asObject).dx := TrackBar1.position;
  TDerivationPaint(splMap.fkt.asObject).dy := TrackBar2.position;
  checkBox2.checked := (TrackBar1.position = 0) and (TrackBar2.position = 0);
  graphSpl.invalidate;
end;

procedure TInterpolation2Dim.ChangeGridVisibility(Sender: TObject);
begin
  bufGrid.visible := checkBox1.checked;
  splGrid.visible := checkBox1.checked;
  graphDat.doPaint(sender);
  graphSpl.doPaint(sender);
end;

procedure TInterpolation2Dim.ChangeSync(Sender: TObject);
begin
  if (checkBox2.checked) and (not splMap.synchronized) then splMap.synchronize(bufMap);
  if (not checkBox2.checked) and (splMap.synchronized) then begin
    splMap.desynchronize;
    splMap.autoColorRange := true;
  end;
  graphSpl.invalidate;
end;

procedure TInterpolation2Dim.delX(Sender: TObject);
var i : integer;
begin
  i := StringGrid1.Selection.left;
  if (i <= 0) then exit;
  StringGrid1.DeleteCol(i);
  tableChange(Sender);
end;

procedure TInterpolation2Dim.delY(Sender: TObject);
var i : integer;
begin
  i := StringGrid1.Selection.top;
  if (i <= 0) then exit;
  StringGrid1.DeleteRow(i);
  tableChange(Sender);
end;

destructor TInterpolation2Dim.destroy;
begin
  buf.destroy;
  if (splMap.fkt <> nil) then splMap.fkt.asObject.destroy;
  inherited destroy;
end;

initialization

TInterpolation2Dim.instanceID := 1;
TGuiTab.addAvaillable(
  '2D interpolation',
  'Features bicubic splines for rectangular grid data.',
  TInterpolation2Dim);

end.

